/*
 * GNOME CPUFreq Applet 
 * Copyright (C) 2004 Carlos Garcia Campos <carlosgc@gnome.org>
 *
 *  This library is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public
 *  License as published by the Free Software Foundation; either
 *  version 2 of the License, or (at your option) any later version.
 *
 *  This library is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public
 *  License along with this library; if not, write to the Free
 *  Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * Authors : Carlos Garca Campos <carlosgc@gnome.org>
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <gtk/gtklabel.h>
#include <gtk/gtkbox.h>
#include <gtk/gtkhbox.h>
#include <gtk/gtkvbox.h>
#include <gtk/gtkimage.h>
#include <gtk/gtkaccessible.h>
#include <gtk/gtkalignment.h>
#include <gtk/gtkaboutdialog.h>
#include <gdk/gdkkeysyms.h>
#include <libgnomeui/gnome-help.h>
#include <panel-applet.h>
#include <panel-applet-gconf.h>
#include <glade/glade.h>
#include <glib/gi18n.h>
#include <string.h>

#include "cpufreq-applet.h"
#include "cpufreq-prefs.h"
#include "cpufreq-popup.h"
#include "cpufreq-monitor.h"
#include "cpufreq-monitor-factory.h"
#include "cpufreq-utils.h"

struct _CPUFreqApplet {
        PanelApplet       base;

        /* Visibility */
        gboolean          show_freq;
        gboolean          show_perc;
        gboolean          show_unit;
        gboolean          show_icon;

        CPUFreqMonitor   *monitor;

        PanelAppletOrient orient;
        gint              size;

        GtkWidget        *label;
        GtkWidget        *unit_label;
        GtkWidget        *icon;
        GtkWidget        *box;
        GtkWidget        *container;
        GdkPixbuf        *pixbufs[5];
        GtkTooltips      *tips;

	gboolean          need_refresh;

        CPUFreqPrefs     *prefs;
        CPUFreqPopup     *popup;
};

struct _CPUFreqAppletClass {
        PanelAppletClass parent_class;
};

static void     cpufreq_applet_init              (CPUFreqApplet      *applet);
static void     cpufreq_applet_class_init        (CPUFreqAppletClass *klass);

static void     cpufreq_applet_preferences_cb    (BonoboUIComponent  *uic,
                                                  CPUFreqApplet      *applet,
                                                  const gchar        *cname);
static void     cpufreq_applet_help_cb           (BonoboUIComponent  *uic,
                                                  CPUFreqApplet      *applet,
                                                  const gchar        *cname);
static void     cpufreq_applet_about_cb          (BonoboUIComponent  *uic,
                                                  CPUFreqApplet      *applet,
                                                  const gchar        *cname);

static void     cpufreq_applet_pixmap_set_image  (CPUFreqApplet      *applet,
                                                  gint                perc);

static void     cpufreq_applet_setup             (CPUFreqApplet      *applet);
static void     cpufreq_applet_update            (CPUFreqApplet      *applet,
                                                  CPUFreqMonitor     *monitor);
static void     cpufreq_applet_refresh           (CPUFreqApplet      *applet);

static void     cpufreq_applet_destroy           (GtkObject          *widget);
static gboolean cpufreq_applet_button_press      (GtkWidget          *widget,
                                                  GdkEventButton     *event);
static gboolean cpufreq_applet_key_press         (GtkWidget          *widget,
                                                  GdkEventKey        *event);
static void     cpufreq_applet_size_allocate     (GtkWidget          *widget,
                                                  GtkAllocation      *allocation);
static void     cpufreq_applet_change_orient     (PanelApplet        *pa,
                                                  PanelAppletOrient   orient);
static void     cpufreq_applet_change_background (PanelApplet        *pa,
                                                  PanelAppletBackgroundType type,
                                                  GdkColor           *color,
                                                  GdkPixmap          *pixmap);

static gboolean cpufreq_applet_factory           (CPUFreqApplet      *applet,
                                                  const gchar        *iid,
                                                  gpointer            gdata);

static const gchar *const cpufreq_icons[] = {
        GNOME_PIXMAPSDIR"/cpufreq-applet/cpufreq-25.png",
        GNOME_PIXMAPSDIR"/cpufreq-applet/cpufreq-50.png",
        GNOME_PIXMAPSDIR"/cpufreq-applet/cpufreq-75.png",
        GNOME_PIXMAPSDIR"/cpufreq-applet/cpufreq-100.png",
        GNOME_PIXMAPSDIR"/cpufreq-applet/cpufreq-na.png",
        NULL
};

static const BonoboUIVerb cpufreq_applet_menu_verbs[] = {
        BONOBO_UI_UNSAFE_VERB ("CPUFreqAppletPreferences",
                               cpufreq_applet_preferences_cb),
        BONOBO_UI_UNSAFE_VERB ("CPUFreqAppletHelp",
                               cpufreq_applet_help_cb),
        BONOBO_UI_UNSAFE_VERB ("CPUFreqAppletAbout",
                               cpufreq_applet_about_cb),
        BONOBO_UI_VERB_END
};

G_DEFINE_TYPE (CPUFreqApplet, cpufreq_applet, PANEL_TYPE_APPLET)

/* Enum Types */
GType
cpufreq_applet_show_mode_get_type (void)
{
        static GType etype = 0;
        
        if (etype == 0) {
                static const GEnumValue values[] = {
                        { CPUFREQ_MODE_GRAPHIC, "CPUFREQ_MODE_GRAPHIC", "mode-graphic" },
                        { CPUFREQ_MODE_TEXT,    "CPUFREQ_MODE_TEXT",    "mode-text" },
                        { CPUFREQ_MODE_BOTH,    "CPUFREQ_MODE_BOTH",    "mode-both" },
                        { 0, NULL, NULL }
                };
                
                etype = g_enum_register_static ("CPUFreqShowMode", values);
        }
        
        return etype;
}

GType
cpufreq_applet_show_text_mode_get_type (void)
{
        static GType etype = 0;

        if (etype == 0) {
                static const GEnumValue values[] = {
                        { CPUFREQ_MODE_TEXT_FREQUENCY,      "CPUFREQ_MODE_TEXT_FREQUENCY",      "mode-text-frequency" },
                        { CPUFREQ_MODE_TEXT_FREQUENCY_UNIT, "CPUFREQ_MODE_TEXT_FREQUENCY_UNIT", "mode-text-frequency-unit" },
                        { CPUFREQ_MODE_TEXT_PERCENTAGE,     "CPUFREQ_MODE_TEXT_PERCENTAGE",     "mode-text-percentage" },
                        { 0, NULL, NULL }
                };

                etype = g_enum_register_static ("CPUFreqShowTextMode", values);
        }

        return etype;
}

GType
cpufreq_applet_selector_mode_get_type (void)
{
        static GType etype = 0;

        if (etype == 0) {
                static const GEnumValue values[] = {
                        { CPUFREQ_SELECTOR_MODE_FREQUENCIES, "CPUFREQ_SELECTOR_MODE_FREQUENCIES", "selector-mode-frequencies" },
                        { CPUFREQ_SELECTOR_MODE_GOVERNORS,   "CPUFREQ_SELECTOR_MODE_GOVERNORS",   "selector-mode-governors" },
                        { CPUFREQ_SELECTOR_MODE_BOTH,        "CPUFREQ_SELECTOR_MODE_BOTH",        "selector-mode-both" },
                        { 0, NULL, NULL }
                };

                etype = g_enum_register_static ("CPUFreqSelectorMode", values);
        }

        return etype;
}

static void
cpufreq_applet_init (CPUFreqApplet *applet)
{
        gint i;

        applet->prefs = NULL;
        applet->popup = NULL;
        applet->monitor = NULL;

        applet->label = gtk_label_new (NULL);
        applet->unit_label = gtk_label_new (NULL);
        applet->icon = gtk_image_new ();
        applet->box = NULL;
           
        applet->container = gtk_alignment_new (0.5, 0.5, 0, 0);
        gtk_container_add (GTK_CONTAINER (applet), applet->container);
        gtk_widget_show (applet->container);
           
        applet->tips = gtk_tooltips_new ();

	applet->need_refresh = TRUE;

        panel_applet_set_flags (PANEL_APPLET (applet), PANEL_APPLET_EXPAND_MINOR);

        applet->size = panel_applet_get_size (PANEL_APPLET (applet));
        applet->orient = panel_applet_get_orient (PANEL_APPLET (applet));

        gtk_widget_show (GTK_WIDGET (applet));
}

static void
cpufreq_applet_class_init (CPUFreqAppletClass *klass)
{
        PanelAppletClass *applet_class = PANEL_APPLET_CLASS (klass);
        GtkObjectClass   *gtkobject_class = GTK_OBJECT_CLASS (klass);
        GtkWidgetClass   *widget_class = GTK_WIDGET_CLASS (klass);

        gtkobject_class->destroy = cpufreq_applet_destroy;
           
        widget_class->size_allocate = cpufreq_applet_size_allocate;
        widget_class->button_press_event = cpufreq_applet_button_press;
        widget_class->key_press_event = cpufreq_applet_key_press;
           
        applet_class->change_orient = cpufreq_applet_change_orient;
        applet_class->change_background = cpufreq_applet_change_background;
}

static void
cpufreq_applet_destroy (GtkObject *widget)
{
        CPUFreqApplet *applet;
        gint           i;

        applet = CPUFREQ_APPLET (widget);

        if (applet->monitor) {
                g_object_unref (G_OBJECT (applet->monitor));
                applet->monitor = NULL;
        }

        if (applet->tips) {
                gtk_object_destroy (GTK_OBJECT (applet->tips));
                applet->tips = NULL;
        }

        for (i = 0; i <= 3; i++) {
                if (applet->pixbufs[i]) {
                        g_object_unref (G_OBJECT (applet->pixbufs[i]));
                        applet->pixbufs[i] = NULL;
                }
        }

        if (applet->prefs) {
                g_object_unref (applet->prefs);
                applet->prefs = NULL;
        }

        if (applet->popup) {
                g_object_unref (applet->popup);
                applet->popup = NULL;
        }

        GTK_OBJECT_CLASS (cpufreq_applet_parent_class)->destroy (widget);
}

static void
cpufreq_applet_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
        CPUFreqApplet *applet;
        gint           size = 0;

        applet = CPUFREQ_APPLET (widget);

        switch (applet->orient) {
        case PANEL_APPLET_ORIENT_LEFT:
        case PANEL_APPLET_ORIENT_RIGHT:
                size = allocation->width;
                break;
        case PANEL_APPLET_ORIENT_UP:
        case PANEL_APPLET_ORIENT_DOWN:
                size = allocation->height;
                break;
        }

        if (size != applet->size) {
                applet->size = size;
                cpufreq_applet_refresh (applet);
        }

        GTK_WIDGET_CLASS (cpufreq_applet_parent_class)->size_allocate (widget, allocation);
}

static void
cpufreq_applet_popup_position_menu (GtkMenu  *menu,
                                    int      *x,
                                    int      *y,
                                    gboolean *push_in,
                                    gpointer  gdata)
{
        GtkWidget      *widget;
        GtkRequisition  requisition;
        gint            menu_xpos;
        gint            menu_ypos;

        widget = GTK_WIDGET (gdata);

        gtk_widget_size_request (GTK_WIDGET (menu), &requisition);

        gdk_window_get_origin (widget->window, &menu_xpos, &menu_ypos);

        menu_xpos += widget->allocation.x;
        menu_ypos += widget->allocation.y;

        switch (panel_applet_get_orient (PANEL_APPLET (widget))) {
        case PANEL_APPLET_ORIENT_DOWN:
        case PANEL_APPLET_ORIENT_UP:
                if (menu_ypos > gdk_screen_get_height (gtk_widget_get_screen (widget)) / 2)
                        menu_ypos -= requisition.height;
                else
                        menu_ypos += widget->allocation.height;
                break;
        case PANEL_APPLET_ORIENT_RIGHT:
        case PANEL_APPLET_ORIENT_LEFT:
                if (menu_xpos > gdk_screen_get_width (gtk_widget_get_screen (widget)) / 2)
                        menu_xpos -= requisition.width;
                else
                        menu_xpos += widget->allocation.width;
                break;
        default:
                g_assert_not_reached ();
        }

        *x = menu_xpos;
        *y = menu_ypos;
        *push_in = TRUE;
}

static void
cpufreq_applet_menu_popup (CPUFreqApplet *applet,
                           guint32        time)
{
        GtkWidget *menu;

        if (!cpufreq_utils_selector_is_available ())
                return;

        if (!applet->popup) {
                applet->popup = cpufreq_popup_new ();
                cpufreq_popup_set_preferences (applet->popup, applet->prefs);
                cpufreq_popup_set_monitor (applet->popup, applet->monitor);
        }

        menu = cpufreq_popup_get_menu (applet->popup);

        if (!menu)
                return;

        gtk_menu_popup (GTK_MENU (menu), NULL, NULL,
                        cpufreq_applet_popup_position_menu,
                        (gpointer) applet,
                        1, time);
}

static gboolean
cpufreq_applet_button_press (GtkWidget *widget, GdkEventButton *event)
{
        CPUFreqApplet *applet;

        applet = CPUFREQ_APPLET (widget);

        if (event->button == 2)
                return FALSE;

        if (event->button == 1 &&
            event->type != GDK_2BUTTON_PRESS &&
            event->type != GDK_3BUTTON_PRESS) {
                cpufreq_applet_menu_popup (applet, event->time);
                
                return TRUE;
        }
        
        return GTK_WIDGET_CLASS (cpufreq_applet_parent_class)->button_press_event (widget, event);
}

static gboolean
cpufreq_applet_key_press (GtkWidget *widget, GdkEventKey *event)
{
        CPUFreqApplet *applet;

        applet = CPUFREQ_APPLET (widget);

        switch (event->keyval) {
        case GDK_KP_Enter:
        case GDK_ISO_Enter:
        case GDK_3270_Enter:
        case GDK_Return:
        case GDK_space:
        case GDK_KP_Space:
                cpufreq_applet_menu_popup (applet, event->time);

                return TRUE;
        default:
                break;
        }

        return FALSE;
}

static void
cpufreq_applet_change_orient (PanelApplet *pa, PanelAppletOrient orient)
{
        CPUFreqApplet *applet;
        gint           size;

        applet = CPUFREQ_APPLET (pa);

        applet->orient = orient;

        if ((orient == PANEL_APPLET_ORIENT_LEFT) ||
            (orient == PANEL_APPLET_ORIENT_RIGHT)) {
                size = GTK_WIDGET (applet)->allocation.width;
        } else {
                size = GTK_WIDGET (applet)->allocation.height;
        }

        if (size != applet->size) {
                applet->size = size;
                cpufreq_applet_refresh (applet);
        }
}

static void
cpufreq_applet_change_background (PanelApplet *pa,
                                  PanelAppletBackgroundType type,
                                  GdkColor *color, GdkPixmap *pixmap)
{
        CPUFreqApplet *applet;
        /* Taken from TrashApplet */
        GtkRcStyle    *rc_style;
        GtkStyle      *style;

        applet = CPUFREQ_APPLET (pa);

        /* reset style */
        gtk_widget_set_style (GTK_WIDGET (applet), NULL);
        rc_style = gtk_rc_style_new ();
        gtk_widget_modify_style (GTK_WIDGET (applet), rc_style);
        gtk_rc_style_unref (rc_style);

        switch (type) {
        case PANEL_PIXMAP_BACKGROUND:
                style = gtk_style_copy (GTK_WIDGET (applet)->style);
                if (style->bg_pixmap[GTK_STATE_NORMAL])
                        g_object_unref (
                                style->bg_pixmap[GTK_STATE_NORMAL]);
                style->bg_pixmap[GTK_STATE_NORMAL] = g_object_ref (
                        pixmap);
                gtk_widget_set_style (GTK_WIDGET (applet), style);
                g_object_unref (style);
                break;
        case PANEL_COLOR_BACKGROUND:
                gtk_widget_modify_bg (GTK_WIDGET (applet),
                                      GTK_STATE_NORMAL, color);
                break;
        case PANEL_NO_BACKGROUND:
        default:
                break;
        }
}

static void
cpufreq_applet_preferences_cb (BonoboUIComponent *uic,
                               CPUFreqApplet     *applet,
                               const gchar       *cname)
{
        cpufreq_preferences_dialog_run (applet->prefs,
                                        gtk_widget_get_screen (GTK_WIDGET (applet)));
}

static void
cpufreq_applet_help_cb (BonoboUIComponent *uic,
                        CPUFreqApplet     *applet,
                        const gchar       *cname)
{
        GError *error = NULL;
           
        gnome_help_display_on_screen ("cpufreq-applet", NULL,
                                      gtk_widget_get_screen (GTK_WIDGET (applet)),
                                      &error);

        if (error) {
                cpufreq_utils_display_error (_("Could not open help document"),
                                             error->message);
                g_error_free (error);
        }
}

static void
cpufreq_applet_about_cb (BonoboUIComponent *uic,
                         CPUFreqApplet     *applet,
                         const gchar       *cname)
{
        static const gchar *const authors[] = {
                "Carlos Garcia Campos <carlosgc@gnome.org>",
                NULL
        };
        static const gchar *const documenters[] = {
                "Carlos Garcia Campos <carlosgc@gnome.org>",
                "Davyd Madeley <davyd@madeley.id.au>",
                NULL
        };
        static const gchar *const artists[] = {
                "Pablo Arroyo Loma <zzioma@yahoo.es>",
                NULL
        };

        gtk_show_about_dialog (NULL,
                               "name",          _("CPU Frequency Scaling Monitor"),
                               "version",       VERSION,
                               "copyright",     "\xC2\xA9 2004 Carlos Garcia Campos",
                               "comments",      _("This utility shows the current CPU "
                                                  "Frequency Scaling."),
                               "authors",       authors,
                               "documenters",   documenters,
                               "artists",       artists, 
                               "translator-credits",    _("translator-credits"),
			       "logo-icon-name",        "gnome-cpu-frequency-applet",
                               NULL);
}

static void
cpufreq_applet_pixmap_set_image (CPUFreqApplet *applet, gint perc)
{
        gint image;

        /* 0-29   -> 25%
         * 30-69  -> 50%
         * 70-89  -> 75%
         * 90-100 -> 100%
         */
        if (perc < 30)
                image = 0;
        else if ((perc >= 30) && (perc < 70))
                image = 1;
        else if ((perc >= 70) && (perc < 90))
                image = 2;
        else if ((perc >= 90) && (perc <= 100))
                image = 3;
        else
                image = 4;

        if (applet->pixbufs[image] == NULL) {
                applet->pixbufs[image] = gdk_pixbuf_new_from_file_at_size (cpufreq_icons[image],
                                                                           24, 24, NULL);
        }

        gtk_image_set_from_pixbuf (GTK_IMAGE (applet->icon), applet->pixbufs[image]);
}

static void
cpufreq_applet_update_visibility (CPUFreqApplet *applet)
{
        CPUFreqShowMode     show_mode;
        CPUFreqShowTextMode show_text_mode;
        gboolean            show_freq = FALSE;
        gboolean            show_perc = FALSE;
        gboolean            show_unit = FALSE;
        gboolean            show_icon = FALSE;
        gboolean            changed = FALSE;
        gboolean            need_update = FALSE;

        show_mode = cpufreq_prefs_get_show_mode (applet->prefs);
        show_text_mode = cpufreq_prefs_get_show_text_mode (applet->prefs);
        
        if (show_mode != CPUFREQ_MODE_GRAPHIC) {
                show_icon = (show_mode == CPUFREQ_MODE_BOTH);

                switch (show_text_mode) {
                case CPUFREQ_MODE_TEXT_FREQUENCY:
                        show_freq = TRUE;
                        break;
                case CPUFREQ_MODE_TEXT_PERCENTAGE:
                        show_perc = TRUE;
                        break;
                case CPUFREQ_MODE_TEXT_FREQUENCY_UNIT:
                        show_freq = TRUE;
                        show_unit = TRUE;
                        break;
                }
        } else {
                show_icon = TRUE;
        }

        if ((applet->show_perc && show_freq) ||
            (applet->show_freq && show_perc))
                need_update = TRUE;
        
        if (show_freq != applet->show_freq) {
                applet->show_freq = show_freq;
                changed = TRUE;
        }

        if (show_perc != applet->show_perc) {
                applet->show_perc = show_perc;
                changed = TRUE;
        }

        if (changed) {
                g_object_set (G_OBJECT (applet->label),
                              "visible",
                              applet->show_freq || applet->show_perc,
                              NULL);
        }

        if (show_unit != applet->show_unit) {
                applet->show_unit = show_unit;
                changed = TRUE;

                g_object_set (G_OBJECT (applet->unit_label),
                              "visible", applet->show_unit,
                              NULL);
        }

        if (show_icon != applet->show_icon) {
                applet->show_icon = show_icon;
                changed = TRUE;

                g_object_set (G_OBJECT (applet->icon),
                              "visible", applet->show_icon,
                              NULL);
        }

        if (changed)
                cpufreq_applet_refresh (applet);

        if (need_update)
                cpufreq_applet_update (applet, applet->monitor);
}

static void
cpufreq_applet_update (CPUFreqApplet *applet, CPUFreqMonitor *monitor)
{
        gchar          *text_tip, *text_mode;
        gchar          *freq_label, *unit_label;
        gint            freq;
        gint            perc;
        guint           cpu;
        gchar          *governor;

        cpu = cpufreq_monitor_get_cpu (monitor);
        freq = cpufreq_monitor_get_frequency (monitor);
        perc = cpufreq_monitor_get_percentage (monitor);
        governor = cpufreq_monitor_get_governor (monitor);

        freq_label = cpufreq_utils_get_frequency_label (freq);
        unit_label = cpufreq_utils_get_frequency_unit (freq);
        
        if (applet->show_freq) {
                gtk_label_set_text (GTK_LABEL (applet->label), freq_label);
        }

        if (applet->show_perc) {
                gchar *text_perc;

                text_perc = g_strdup_printf ("%d%%", perc);
                gtk_label_set_text (GTK_LABEL (applet->label), text_perc);
                g_free (text_perc);
        }

        if (applet->show_unit) {
                gtk_label_set_text (GTK_LABEL (applet->unit_label), unit_label);
        }

        if (applet->show_icon) {
                cpufreq_applet_pixmap_set_image (applet, perc);
        }

        governor[0] = g_ascii_toupper (governor[0]);
        text_mode = g_strdup_printf ("%s\n%s %s (%d%%)",
                                     governor, freq_label,
                                     unit_label, perc);

        g_free (freq_label);
        g_free (unit_label);
        g_free (governor);

        if (cpufreq_utils_get_n_cpus () == 1)
                text_tip = g_strdup_printf ("%s", text_mode);
        else
                text_tip = g_strdup_printf ("CPU %u - %s", cpu, text_mode);

        g_free (text_mode);
           
        gtk_tooltips_set_tip (applet->tips, GTK_WIDGET (applet), text_tip, NULL);
        g_free (text_tip);

        /* Call refresh only the first time */
        if (applet->need_refresh) {
                cpufreq_applet_refresh (applet);
		applet->need_refresh = FALSE;
        }
}

static gint
cpufreq_applet_get_widget_size (CPUFreqApplet *applet,
                                GtkWidget     *widget)
{
        GtkRequisition  req;
        gint            size;

        if (!GTK_WIDGET_VISIBLE (widget))
                return 0;
        
        gtk_widget_size_request (widget, &req);
        
        switch (applet->orient) {
        case PANEL_APPLET_ORIENT_LEFT:
        case PANEL_APPLET_ORIENT_RIGHT:
                size = req.width;
                break;
        case PANEL_APPLET_ORIENT_UP:
        case PANEL_APPLET_ORIENT_DOWN:
                size = req.height;
                break;
        default:
                g_assert_not_reached ();
        }

        return size;
}

static void
cpufreq_applet_refresh (CPUFreqApplet *applet)
{
        GtkWidget *labels_box = NULL;
        gint       total_size = 0;
        gint       panel_size, label_size;
        gint       unit_label_size, pixmap_size;
        gint       size_step = 12;
        gboolean   do_unref = FALSE;

        panel_size = applet->size - 1; /* 1 pixel margin */

        /* We want a fixed label size, the biggest */
        gtk_label_set_width_chars (GTK_LABEL (applet->label), 4);
        label_size = cpufreq_applet_get_widget_size (applet, applet->label);
        gtk_label_set_width_chars (GTK_LABEL (applet->label), -1);
        total_size += label_size;
           
        unit_label_size = cpufreq_applet_get_widget_size (applet, applet->unit_label);
        total_size += unit_label_size;
        
        pixmap_size = cpufreq_applet_get_widget_size (applet, applet->icon);
        total_size += pixmap_size;

        if (applet->box) {
                GtkWidget *child;
                GList     *children;

                do_unref = TRUE;
                g_object_ref (applet->icon);
                gtk_container_remove (GTK_CONTAINER (applet->box), applet->icon);
                children = gtk_container_get_children (GTK_CONTAINER (applet->box));
                if (children && children->data) {
                        /* Should be labels_box */
                        child = (GtkWidget *)children->data;
                        g_object_ref (applet->label);
                        gtk_container_remove (GTK_CONTAINER (child), applet->label);
                        g_object_ref (applet->unit_label);
                        gtk_container_remove (GTK_CONTAINER (child), applet->unit_label);
                }
                g_list_free (children);
                gtk_widget_destroy (applet->box);
        }

        switch (applet->orient) {
        case PANEL_APPLET_ORIENT_LEFT:
        case PANEL_APPLET_ORIENT_RIGHT:
                if (total_size <= panel_size) {
                        applet->box = gtk_hbox_new (FALSE, 2);
                        labels_box  = gtk_hbox_new (FALSE, 2);
                } else if ((label_size + unit_label_size) <= (panel_size - size_step)) {
                        applet->box = gtk_vbox_new (FALSE, 2);
                        labels_box  = gtk_hbox_new (FALSE, 2);
                } else {
                        applet->box = gtk_vbox_new (FALSE, 2);
                        labels_box  = gtk_vbox_new (FALSE, 2);
                }
                break;
        case PANEL_APPLET_ORIENT_UP:
        case PANEL_APPLET_ORIENT_DOWN:
                labels_box = gtk_hbox_new (FALSE, 2);
                if ((label_size + pixmap_size) <= panel_size)
                        applet->box = gtk_vbox_new (FALSE, 2);
                else
                        applet->box = gtk_hbox_new (FALSE, 2);
                break;
        default:
                g_assert_not_reached ();
        }

        gtk_box_pack_start (GTK_BOX (labels_box), applet->label, FALSE, TRUE, 0);
        gtk_box_pack_start (GTK_BOX (labels_box), applet->unit_label, FALSE, TRUE, 0);
        gtk_box_pack_start (GTK_BOX (applet->box), applet->icon, FALSE, TRUE, 0);
        
        gtk_box_pack_start (GTK_BOX (applet->box), labels_box, FALSE, TRUE, 0);
        gtk_widget_show (labels_box);
           
        gtk_container_add (GTK_CONTAINER (applet->container), applet->box);
        gtk_widget_show (applet->box);

        if (do_unref) {
                g_object_unref (applet->label);
                g_object_unref (applet->unit_label);
                g_object_unref (applet->icon);
        }
}

/* Preferences callbacks */
static void
cpufreq_applet_prefs_cpu_changed (CPUFreqPrefs  *prefs,
				  GParamSpec    *arg1,
				  CPUFreqApplet *applet)
{
	cpufreq_monitor_set_cpu (applet->monitor,
				 cpufreq_prefs_get_cpu (applet->prefs));
}

static void
cpufreq_applet_prefs_show_mode_changed (CPUFreqPrefs  *prefs,
                                        GParamSpec    *arg1,
                                        CPUFreqApplet *applet)
{
        cpufreq_applet_update_visibility (applet);
}

static void
cpufreq_applet_setup (CPUFreqApplet *applet)
{
        BonoboUIComponent *popup_component;
        AtkObject         *atk_obj;
        gchar             *prefs_key;

        glade_gnome_init ();

	gtk_window_set_default_icon_name ("gnome-cpu-frequency-applet");
	
        panel_applet_add_preferences (PANEL_APPLET (applet),
                                      "/schemas/apps/cpufreq-applet/prefs", NULL);
           
        /* Preferences */
        if (applet->prefs)
                g_object_unref (applet->prefs);
        
        prefs_key = panel_applet_get_preferences_key (PANEL_APPLET (applet));
        applet->prefs = cpufreq_prefs_new (prefs_key);
        g_free (prefs_key);

	g_signal_connect (G_OBJECT (applet->prefs),
			  "notify::cpu",
			  G_CALLBACK (cpufreq_applet_prefs_cpu_changed),
			  (gpointer) applet);
        g_signal_connect (G_OBJECT (applet->prefs),
                          "notify::show-mode",
                          G_CALLBACK (cpufreq_applet_prefs_show_mode_changed),
                          (gpointer) applet);
        g_signal_connect (G_OBJECT (applet->prefs),
                          "notify::show-text-mode",
                          G_CALLBACK (cpufreq_applet_prefs_show_mode_changed),
                          (gpointer) applet);
        
        cpufreq_applet_update_visibility (applet);

        /* Monitor */
        applet->monitor = cpufreq_monitor_factory_create_monitor (
                cpufreq_prefs_get_cpu (applet->prefs));
        cpufreq_monitor_run (applet->monitor);
        g_signal_connect_swapped (G_OBJECT (applet->monitor), "changed",
                                  G_CALLBACK (cpufreq_applet_update),
                                  (gpointer) applet);
           
        /* Setup the menus */
        panel_applet_setup_menu_from_file (PANEL_APPLET (applet),
                                           DATADIR,
                                           "GNOME_CPUFreqApplet.xml",
                                           NULL,
                                           cpufreq_applet_menu_verbs,
                                           applet);

        if (panel_applet_get_locked_down (PANEL_APPLET (applet))) {
                popup_component = panel_applet_get_popup_component (PANEL_APPLET (applet));
                         
                bonobo_ui_component_set_prop (popup_component,
                                              "/commands/CPUFreqPreferences",
                                              "hidden", "1",
                                              NULL);
        }

        atk_obj = gtk_widget_get_accessible (GTK_WIDGET (applet));

        if (GTK_IS_ACCESSIBLE (atk_obj)) {
                atk_object_set_name (atk_obj, _("CPU Frequency Scaling Monitor"));
                atk_object_set_description (atk_obj, _("This utility shows the current CPU Frequency"));
        }
}

static gboolean
cpufreq_applet_factory (CPUFreqApplet *applet, const gchar *iid, gpointer gdata)
{
        gboolean retval = FALSE;

        if (!strcmp (iid, "OAFIID:GNOME_CPUFreqApplet")) {
                cpufreq_applet_setup (applet);
                
                retval = TRUE;
        }

        return retval;
}

PANEL_APPLET_BONOBO_FACTORY ("OAFIID:GNOME_CPUFreqApplet_Factory",
                             CPUFREQ_TYPE_APPLET,
                             "cpufreq-applet",
                             "0",
                             (PanelAppletFactoryCallback) cpufreq_applet_factory,
                             NULL)
